/*
Murasaki - multiple genome global alignment program
Copyright (C) 2006-2007 Keio University
(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef __DINKYMATH_H
#define __DINKYMATH_H

#include <iostream>
#include <vector>
#include <set>
#include <iterator>

using namespace std;

//int log2(int x);
inline int log2(int b){
  int l=0;
  while(b>0){
    b>>=1;
    l++;
  }
  return l;
}

inline int intpow(int base,int pow){
  int b=base;
  for(int i=1;i<pow;i++)
    base*=b;
  return base;
}

template <class T>
inline T sqr(const T& a){
  return a*a;
}

//int intpow(int base,int pow);
inline int abs(int a){return a<0 ? -a:a;}

string littleEndianStr(unsigned int n,int digits);
string littleEndianPat(unsigned int n,int digits);
string littleEndianStr(unsigned int n);

template <class T>
int writeOutVector(string filename,const vector<T> &v,string delim,int,int,int skipZeros=0);

string dstring(int);
string fstring(float);
string humanMemory(unsigned long bytes);
string humanMemory(unsigned int bytes);
string humanMemory(int bytes);
string humanMemory(double bytes);
string percent(double,double);
int progressTick(int i,int max);
void writeOut(string filename,string content);

class RandomTraverse{
 public:
  int next();
  int remaining();
  RandomTraverse(int min,int max);
 protected:
  int min,max;
  set<int> hit;
};

class ProgressTicker {
 public:
  ProgressTicker(int _max);
  int tick(int i);
  inline int tick(){return tick(lastCall+1);}
  void reset(int _max);
 protected:
  int max;
  int lineDelay,dotDelay,lineCount,dotCount,lastCall;
};

unsigned gcd(unsigned a,unsigned b);
unsigned lcm(unsigned a,unsigned b);
#endif

