#include "ecolist.h"
#include "dinkymath.h"
#include <assert.h>
#include <algorithm>

bool Ecolist::inited=false;
int Ecolist::minBlocks,Ecolist::minChunks;
unsigned Ecolist::seqbits,Ecolist::idxbits,Ecolist::chunkSize;
word Ecolist::idxLongest;
int Ecolist::initBlocks;
int Ecolist::externalSortLimit=1<<28; //this is about 2.68435e8
bool (*Ecolist::lessthan)(const pair<int,int>&,const pair<int,int>&)=0;

Ecolist::Ecolist() :
  size(0)
{
  assert(inited);
  blocks=allocBlob(0);
  assert(blocks);
}

void Ecolist::clear(){
  eraseAll(blocks,0);
  size=0;
  blocks=allocBlob(0);
}

ostream& operator<<(ostream& os,Ecolist& a){
  return os << "(Ecolist: "<<a.getSize()<<"/"<<a.capacity()<<")";
}

void Ecolist::push(int seq,int idx){
  int loc=size++;
  block* root=findBlockFor(blocks,loc);
  writeBits(root,loc,0,seqbits,seq);
  writeBits(root,loc,seqbits,idxbits,idx+idxLongest);
}

std::pair<int,int> Ecolist::getValAt(int loc){
  assert(loc<size);
  block* root=findBlockFor(blocks,loc);
  int seq=readBits(root,loc,0,seqbits);
  int idx=readBits(root,loc,seqbits,idxbits)-idxLongest;
  return std::pair<int,int>(seq,idx);
}

int Ecolist::capacity(){
  int cap=0;
  int ite=0;
  for(block *root=blocks;root;root=root[initBlocks<<ite++].p)
    cap+=minChunks<<ite;
  return cap;
}

Ecolist::iterator Ecolist::begin(){
  return iterator(blocks,0,0);
}

Ecolist::iterator Ecolist::end(){
  int ite=0;
  block *root;
  int pos=size;
  for(root=blocks;root;){
    block *tmp=root[initBlocks<<ite].p;
    if(tmp){
      pos-=minChunks<<ite++;
      root=tmp;
    }
    else
      break;
  }
  return iterator(root,pos,ite);
}

Ecolist::iterator Ecolist::at(int remain){
  if(remain>=size)
    return end();
  int ite=0;
  block* root=blocks;
  while(remain>=minChunks<<ite){
    assert(root[initBlocks<<ite].p);
    root=root[initBlocks<<ite].p;
    remain-=minChunks<<ite;
    ite++;
  }
  return iterator(root,remain,ite);
}

//bidirectional versions
Ecolist::bi_iterator Ecolist::bi_begin(){
  return bi_iterator(blocks,list<block*>(),0,0);
}

Ecolist::bi_iterator Ecolist::bi_end(){
  int ite=0;
  list<block*> prev;
  block *root;
  int pos=size;
  for(root=blocks;root;){
    block *tmp=root[initBlocks<<ite].p;
    if(tmp){
      prev.push_back(root);
      pos-=minChunks<<ite++;
    }else
      break;
    root=tmp;
  }
  return bi_iterator(root,prev,pos,ite);
}

Ecolist::bi_iterator Ecolist::bi_at(int remain){
  if(remain>=size)
    return bi_end();
  int ite=0;
  list<block*> prev;
  block* root=blocks;
  while(remain>=minChunks<<ite){
    prev.push_back(root);
    assert(root[initBlocks<<ite].p);
    root=root[initBlocks<<ite].p;
    remain-=minChunks<<ite;
    ite++;
  }
  return bi_iterator(root,prev,remain,ite);
}

//static mojo
void Ecolist::init(int seqs,word longest,word total,int weight,bool (*_lt)(const val_type&,const val_type&))
{
  idxLongest=longest;
  seqbits=log2(seqs);
  idxbits=log2((longest+1)*2);
  chunkSize=seqbits+idxbits;
  minBlocks=lcm(log2(seqs)+log2((longest+1)*2),WORDSIZE)/WORDSIZE;
  int expected=log2(total)-weight;
  std::cerr << "Expcted: "<<expected<<endl;
  for(initBlocks=minBlocks;initBlocks<expected;initBlocks+=minBlocks);
  minChunks=WORDSIZE*initBlocks/(seqbits+idxbits);
  std::cerr << "minChunks: "<<minChunks<<endl;
  std::cerr << "(Ecolist~ seqbits: "<<seqbits<<" idxbits: "<<idxbits<<" minBlocks: "<<minBlocks<<" initblocks: "<<initBlocks<<")"<<endl;
  inited=true;
  lessthan=_lt;
}

//protected mojo
Ecolist::~Ecolist(){
  eraseAll(blocks,0);
}

void Ecolist::writeBits(block* root,int loc,int offset,int size,int val){
  int bitloc=loc*chunkSize+offset;
  int wloc=bitloc/WORDSIZE;
  offset=MODWORDSIZE(bitloc); //offset now means "location inside word to start writing from"
  word mask=lowN(size)<<offset;
  root[wloc].w=(root[wloc].w & (~mask)) | ((word)(val & lowN(size))<<offset);
  if(offset && MODWORDSIZE(bitloc+size-1)<MODWORDSIZE(bitloc-1)){ //overlaps into next word
    int valoff=WORDSIZE-offset; //first bit of val not yet written
    size-=valoff; //size is now the remainder. had better fit in 1 word!
    assert(size>0);
    assert(size<WORDSIZE);
    bitloc+=valoff;
    assert(MODWORDSIZE(bitloc)==0); //bitloc should now be exactly at the word-boundary
    assert(bitloc/WORDSIZE==wloc+1); //and that word should be the one right after wloc
    wloc++;
    //offset=0; //we can simplify the operation just by leaving this out
    mask=lowN(size);
    root[wloc].w=(root[wloc].w & (~mask)) | ((word)(val & (mask<<valoff))>>valoff);
  }
}

int Ecolist::readBits(block* root,int loc,int offset,int size){
  int bitloc=loc*chunkSize+offset;
  int wloc=bitloc/WORDSIZE;
  offset=MODWORDSIZE(bitloc); //offset now means "location inside word to start reading from"
  word mask=lowN(size)<<offset;
  word res=(root[wloc].w & mask)>>offset;
  if(offset && MODWORDSIZE(bitloc+size-1)<MODWORDSIZE(bitloc-1)){ //overlaps into next word
    int valoff=WORDSIZE-offset; //first bit of val not yet written
    size-=valoff; //size is now the remainder. had better fit in 1 word!
    assert(size>0);
    assert(size<WORDSIZE);
    bitloc+=valoff;
    assert(MODWORDSIZE(bitloc)==0); //bitloc should now be exactly at the word-boundary
    assert(bitloc/WORDSIZE==wloc+1); //and that word should be the one right after wloc
    wloc++;
    //offset=0; //we can simplify the operation just by leaving this out
    mask=lowN(size);
    res=res | (word)(root[wloc].w & mask)<<valoff;
  }
  return (int)res;
}

void Ecolist::eraseAll(block* root,int ite){
  block* next=root[initBlocks<<ite].p;
  delete root;
  if(next)
    eraseAll(next,ite+1);
}

block* Ecolist::findBlockFor(block* root,int &remain){
  int ite=0;
  while(remain>=minChunks<<ite){
    //get next root (create if necessary), and dec remain
    block*& next=root[initBlocks<<ite].p;
    if(!next)
      next=allocBlob(ite+1);
    root=next;
    remain-=minChunks<<ite;
    ite++;
  }
  return root;
}

block* Ecolist::allocBlob(int ite){
  //  std::cerr << "Allocing new blob for level "<<ite<<endl;
  int last=initBlocks<<ite;
  block* n=new block[last+1];
  n[last].p=0;
  return n;
}

//iterator mojo
Ecolist::EcolistIterator::EcolistIterator(block *_start,int _p,int _ite) :
  current(_start),pos(_p),ite(_ite),end(Ecolist::minChunks<<_ite)
{ }

Ecolist::EcolistIterator::EcolistIterator(const Ecolist::bi_iterator &a) :
  current(a.current),pos(a.pos),ite(a.ite),end(a.end)
{ }

std::pair<int,int> Ecolist::EcolistIterator::operator*() const {
  int seq=readBits(current,pos,0,Ecolist::seqbits);
  int idx=readBits(current,pos,Ecolist::seqbits,Ecolist::idxbits)-Ecolist::idxLongest;
  return std::pair<int,int>(seq,idx);
}

std::pair<int,int> Ecolist::EcolistIterator::get() const {
  int seq=readBits(current,pos,0,Ecolist::seqbits);
  int idx=readBits(current,pos,Ecolist::seqbits,Ecolist::idxbits)-Ecolist::idxLongest;
  return std::pair<int,int>(seq,idx);
}

Ecolist::EcolistIterator::_Self& Ecolist::EcolistIterator::operator++(){
  pos++;
  if(pos==end && current[initBlocks<<ite].p){ //roll into next block
    assert(current[initBlocks<<ite].p); //better not be null!
    current=current[initBlocks<<ite].p;
    pos=0;
    end=Ecolist::minChunks<<++ite;
  }
  return *this;
}

Ecolist::EcolistIterator::_Self Ecolist::EcolistIterator::operator++(int a){
  _Self __tmp=*this;
  pos++;
  if(pos==end && current[initBlocks<<ite].p){ //roll into next block
    assert(current[initBlocks<<ite].p); //better not be null!
    current=current[initBlocks<<ite].p;
    pos=0;
    end=Ecolist::minChunks<<++ite;
  }
  return __tmp;
}

bool Ecolist::EcolistIterator::operator!=(const _Self& a) const {
  return current!=a.current || pos!=a.pos;
}

bool Ecolist::EcolistIterator::operator==(const _Self& a) const {
  return current==a.current && pos==a.pos;
}

void Ecolist::EcolistIterator::put(int seq,int idx) const {
  writeBits(current,pos,0,Ecolist::seqbits,seq);
  writeBits(current,pos,Ecolist::seqbits,Ecolist::idxbits,idx+Ecolist::idxLongest);
}

void Ecolist::EcolistIterator::put(const pair<int,int> &v) const {
  writeBits(current,pos,0,Ecolist::seqbits,v.first);
  writeBits(current,pos,Ecolist::seqbits,Ecolist::idxbits,v.second+Ecolist::idxLongest);
}

//bidirectional-mojo
Ecolist::BidirectionalEcolistIterator::BidirectionalEcolistIterator(block *_start,list<block*> _prev,int _p,int _ite) :
  current(_start),prev(_prev),pos(_p),ite(_ite),end(Ecolist::minChunks<<_ite)
{ }

std::pair<int,int> Ecolist::BidirectionalEcolistIterator::operator*() const {
  int seq=readBits(current,pos,0,Ecolist::seqbits);
  int idx=readBits(current,pos,Ecolist::seqbits,Ecolist::idxbits)-Ecolist::idxLongest;
  return std::pair<int,int>(seq,idx);
}

std::pair<int,int> Ecolist::BidirectionalEcolistIterator::get() const {
  int seq=readBits(current,pos,0,Ecolist::seqbits);
  int idx=readBits(current,pos,Ecolist::seqbits,Ecolist::idxbits)-Ecolist::idxLongest;
  return std::pair<int,int>(seq,idx);
}

Ecolist::BidirectionalEcolistIterator::_Self& Ecolist::BidirectionalEcolistIterator::operator++(){
  pos++;
  if(pos==end && current[initBlocks<<ite].p){ //roll into next block
    assert(current[initBlocks<<ite].p); //better not be null!
    prev.push_back(current);
    current=current[initBlocks<<ite].p;
    pos=0;
    end=Ecolist::minChunks<<++ite;
  }
  return *this;
}

Ecolist::BidirectionalEcolistIterator::_Self Ecolist::BidirectionalEcolistIterator::operator++(int a){
  _Self __tmp=*this;
  pos++;
  if(pos==end && current[initBlocks<<ite].p){ //roll into next block
    assert(current[initBlocks<<ite].p); //better not be null!
    prev.push_back(current);
    current=current[initBlocks<<ite].p;
    pos=0;
    end=Ecolist::minChunks<<++ite;
  }
  return __tmp;
}

Ecolist::BidirectionalEcolistIterator::_Self& Ecolist::BidirectionalEcolistIterator::operator--(){
  if(pos==0){ //roll into prev block
    assert(!prev.empty()); //better not be null!
    current=prev.back();
    prev.pop_back();
    pos=end=Ecolist::minChunks<<--ite;
  }
  pos--;
  return *this;
}

Ecolist::BidirectionalEcolistIterator::_Self Ecolist::BidirectionalEcolistIterator::operator--(int a){
  _Self __tmp=*this;
  if(pos==0){ //roll into prev block
    assert(!prev.empty()); //better not be null!
    current=prev.back();
    prev.pop_back();
    pos=end=Ecolist::minChunks<<--ite;
  }
  pos--;
  return __tmp;
}

bool Ecolist::BidirectionalEcolistIterator::operator!=(const _Self& a) const {
  return current!=a.current || pos!=a.pos;
}

bool Ecolist::BidirectionalEcolistIterator::operator==(const _Self& a) const {
  return current==a.current && pos==a.pos;
}

void Ecolist::BidirectionalEcolistIterator::put(int seq,int idx) const {
  writeBits(current,pos,0,Ecolist::seqbits,seq);
  writeBits(current,pos,Ecolist::seqbits,Ecolist::idxbits,idx+Ecolist::idxLongest);
}

void Ecolist::BidirectionalEcolistIterator::put(const pair<int,int> &v) const {
  writeBits(current,pos,0,Ecolist::seqbits,v.first);
  writeBits(current,pos,Ecolist::seqbits,Ecolist::idxbits,v.second+Ecolist::idxLongest);
}

//sort mojo
void Ecolist::inPlaceSort(){
  assert(lessthan);
  qsort(0,size);
}

int Ecolist::parition(int left,int right,int pivoti){
  iterator pi(at(pivoti)),li(at(left)),ri(at(right));
  val_type pv=*pi;
  swap(pi,ri);
  iterator store(li);int si=left; //using a store lets us get away with forward-only iterators
  for(iterator i(li),stop(ri);i!=stop;++i){
    if(lessthan(*i,pv)){
      swap(store,i);
      ++store;++si;
    }
  }
  swap(ri,store);
  return si;
}

void Ecolist::qsort(int left,int right){
  if(right<=left)
    return;
  int pivoti=(left+right)/2; //middle. perhaps we should be more clever? but i doubt it.
  pivoti=parition(left,right,pivoti);
  qsort(left,pivoti-1);
  qsort(pivoti+1,right);
}

void Ecolist::swap(iterator &a,iterator &b){
  val_type temp=*a;
  a.put(*b);
  b.put(temp);
}

struct less_map : public binary_function<const Ecolist::val_type&,const Ecolist::val_type&,bool> {
  bool operator()(const Ecolist::val_type& a,const Ecolist::val_type& b) { return Ecolist::lessthan(a,b); }
};

void Ecolist::externalSort(){
  vector<val_type> v;
  v.reserve(size);
  iterator ite(begin()),stop(end());
  for(;ite!=stop;++ite)
    v.push_back(*ite);
  std::sort(v.begin(),v.end(),less_map());
  ite=begin();
  for(int i=0;i<size;i++,ite++)
    ite.put(v[i]);
}

int Ecolist::int_lower_bound(const val_type &m){ //first elem not <m
  int low=0,high=size;
  int len=high-low;
  while(len>0){
    int half=len/2;
    int mid=low+half;
    high=low+len;
    val_type here=getValAt(mid);
    if(lessthan(here,m)){
      low=mid+1;
      len=len-half-1;
    }
    else
      len=half;
  }
  return low;
}

int Ecolist::int_upper_bound(const val_type &m){ //first elem not <m
  int low=0,high=size;
  int len=high-low;
  while(len>0){
    int half=len/2;
    int mid=low+half;
    high=low+len;
    val_type here=getValAt(mid);
    if(lessthan(m,here))
      len=half;
    else{
      low=mid+1;
      len=len-half-1;
    }
  }
  return low;
}
